WITH Text_Io; USE Text_Io;

WITH Ada.strings.Unbounded, comtypes;
use Ada.Strings.Unbounded, comtypes;

package Baumwandler_Package is

-----------------------------------------------------------------------------------------------
--------- Typdeklarationen --------------------------------------------------------------------
-- Konstruktion der Syntax der while-Sprache --------------------------------------------------
-- Einfhrung aller notwendigen Datentypen ----------------------------------------------------
-----------------------------------------------------------------------------------------------
   type Operator is (Plus,Minus,Mal,Div,Modulo);
-----------------------------------------------------------------------------------------------
   type Vergleicher is (Kl,Gl,Gr,Kg,Ug,Gg);
-----------------------------------------------------------------------------------------------
   type Ausdrucks_Art is (Zahl,Name,Verkn);
   type Ausdruck_Rec;
   type Ausdruck is access Ausdruck_Rec;
   type Ausdruck_Rec (Form : Ausdrucks_Art := Zahl) is record
      case Form is
         when Zahl  => Zahl_Wert       : Integer;

         when Name  => Name_Wert       : Unbounded_String;

         when Verkn => Operator_Wert   : Operator;
                       Links_Ausdruck,
                       Rechts_Ausdruck : Ausdruck;
      end case;
   end record;
-----------------------------------------------------------------------------------------------
   type Bedingungs_Art is (Wahr,Vergleich);
   type Bedingung_Rec;
   type Bedingung is access Bedingung_Rec;
   type Bedingung_Rec (Form : Bedingungs_Art := Wahr) is record
      case Form is
         when Wahr      => Wahr_Wert       : Boolean;

         when Vergleich => Vergleichs_Wert : Vergleicher;
                           Links_Ausdruck,
                           Rechts_Ausdruck : Ausdruck;
      end case;
   end record;
-----------------------------------------------------------------------------------------------
   type Anweisungs_Art is (Fall,Schleife,Skip,Zuweisung);
   type Anweisung_Rec;
   type Anweisung_F_Rec;
   type Anweisung is access Anweisung_Rec;
   type Anweisung_F is access Anweisung_F_Rec;
   type Anweisung_F_Rec is record
         Anw  : Anweisung;
         Rest : Anweisung_F;
   end record;
   type Anweisung_Rec (Form : Anweisungs_Art := Fall) is record
      case Form is
         when Fall      => Bedingungs_Wert_Fall     : Bedingung;
                           Erst_Anweisung_Fall,
                           Zweit_Anweisung_Fall     : Anweisung_F;

         when Schleife  => Bedingungs_Wert_Schleife : Bedingung;
                           Erst_Anweisung_Schleife  : Anweisung_F;

         when Skip      => null;

         when Zuweisung => Zuweisungs_Name          : Unbounded_String;
                           Links_Ausdruck           : Ausdruck;
      end case;
   end record;

-- Da Anweisungen mehrfach auch Listen von Anweisungen enthalten knnen,
-- muss der Typ der Anweisungsfolge extra deklariert werden. Die Typen
-- Anweisung und Anweisung_F werden letztlich auch gegenseitig zur Dekla-
-- ration verwendet. Dies ist ein Beispiel von vielen, in der der Mehr-
-- aufwand in Ada gegenber der Haskellversion sehr deutlich wird.
-----------------------------------------------------------------------------------------------
   type Wert_Art is (Undef,Def_Wert);
   type Wert (Form : Wert_Art := Undef) is record
      case Form is
         when Undef    => null;
         when Def_Wert => Zahl_Wert : Integer;
      end case;
   end record;

-- Speicherbelgung einer Variable in der Umgebung
-----------------------------------------------------------------------------------------------
   type Variable is record
         Variable_Name : Unbounded_String;
         Variable_Wert : Wert;
   end record;

-- Variable als Tupel des Variablennamens und der Speicherbelegung
-----------------------------------------------------------------------------------------------
   type Variable_Rec;
   type Variablen is access Variable_Rec;
   type Variable_Rec is record
      Variable_Name : Unbounded_String;
      Variable_Wert : Wert;
      Variable_Next : Variablen;
   end record;
-----------------------------------------------------------------------------------------------
   type Umgebung_Rec;
   type Umgebung is access Umgebung_Rec;
   type Umgebung_Rec is record
      Elem : Variable;
      Next : Umgebung;
   end record;

-- Die Umgebung wird aus einer Liste von Variablen erzeugt. Jedes
-- Listenelement enthlt dabei Variablennamen und Speicherbelegung.
-----------------------------------------------------------------------------------------------
   type Teil_Arten is (Anw,Anw_F,Bed,Ausdr,Bez,Zwert,Wwert,Op,Rop,If_Case,Do_Case,Zuw);
   type Teil_Rec;
   type Teil is access Teil_Rec;
   type Teil_Rec (Form : Teil_Arten := Anw) is record
      case Form is
         when Anw     => Teil_Anw   : Anweisung;

         when Anw_F   => Teil_Anw_F : Anweisung_F;

         when Bed     => Teil_Bed   : Bedingung;

         when Ausdr   => Teil_Ausdr : Ausdruck;

         when Bez     => Teil_Bez   : Unbounded_String;

         when Zwert   => Teil_Zwert : Integer;

         when Wwert   => Teil_Wwert : Boolean;

         when Op      => Teil_Op    : Operator;

         when Rop     => Teil_Rop   : Vergleicher;

         when others  => null;
      end case;
   end record;

-- Der Datentyp Teil definiert den Aufbau des Programmkellers und
-- des Wertekellers. Zwar wird nicht jedes mgliche Programmteil
-- bei der Auswertung auch im Wertekeller gespeichert, jedoch ist
-- es bersichtlicher nur einen Typen fr beide Keller zu nutzen.
-----------------------------------------------------------------------------------------------
   type Teile_Rec;
   type Teile is access Teile_Rec;
   type Teile_Rec is record
      T    : Teil;
      Rest : Teile;
   end record;
-----------------------------------------------------------------------------------------------
   type Konfiguration is record
      Um  : Umgebung;
      Wk,
      Zk  : Teile;
   end record;

-- Der Zustand der abstrakten Maschine als Tripple der Speicher-
-- umgebung, dem Wertekeller und dem Programmkeller
-----------------------------------------------------------------------------------------------
   type Zustand_Arten is (Ende, Akt);
   type Zustand (Form : Zustand_Arten := Ende) is record
      case Form is
         when Ende => Ende_Zustand : Umgebung;

         when Akt  => Akt_Zustand  : Konfiguration;
      end case;
   end record;

   WhileInit_Error : exception;
   
-- Der Datentyp Zustand, um vom Anfangszustand sowie den
-- den Zwischenzustnden whrend der Auswertung und von
-- der Endspeicherbelegung, fr die die Kellerinhalte nicht
-- mehr von Bedeutung sind, unterscheiden zu knnen.
--------------------------------------------------------------------------------------
--------------------------------------------------------------------------------------

   procedure macheTrace (Baum: GrBaum);
   procedure macheTrace (Baum: GrBaum; FT: File_Type);

   private

--   procedure Trace_File (Z: Zustand; FT: File_Type);
   procedure Trace (Z: Zustand);
   function opt_Wandler (op: Unbounded_String) return Operator;
   function Ausdruck_Wandler (Baum: GrBaum) return Ausdruck;
   function Ropt_Wandler (Rop: Unbounded_String) return Vergleicher;
   function Bedingung_Wandler (Baum: GrBaum) return Bedingung;
   function Anw_Wandler (Baum: GrBaum) return Anweisung;
   function Anwf_Wandler (Baum: Grbaum) return Anweisung_F;
   function Baum_Wandler (Baum: Grbaum) return Teile;

   Function Wert_Suche (Var: Unbounded_String; Umg: Umgebung) return Boolean;
   Function Var_Tupel (gr:GrBaum; Var_Liste: Umgebung) return Umgebung;


-----------------------------------------------------------------------------------------------
--------- Funktionen: kopiert aus While-Interpreter -------------------------------------------
-----------------------------------------------------------------------------------------------

   function Ada_Op (Op: Operator; Z1, Z2: Integer) return Integer;
   function Ada_Rel_Op (Rop: Vergleicher; Z1, Z2: Integer) return Boolean;
   function Ist_Undef (W : Wert) return Boolean;
   function VarD return Variable;
   function Wert_Suche (Var : Unbounded_String; Umg : Umgebung) return Wert;
   function Wert_Eintrag (Var: Unbounded_String; Int_Wert: Integer; Umg: Umgebung) return Umgebung;

-----------------------------------------------------------------------------------------------
---------- Zeige-Funktionen: kopiert aus While-Interpreter ------------------------------------
-----------------------------------------------------------------------------------------------

   function Zeige_Op (A : Operator) return String;
   function Zeige_Vergl (A : Vergleicher ) return String;
   function Zeige_Ausdr (A : Ausdruck) return String;
   function Zeige_Bed (A : Bedingung) return String;
   function Zeige_Anw_F (A : Anweisung_F) return String;
   function Zeige_Teile (A : Teile) return String;
   function Zeige_Wert (A : Wert) return String;
   function Zeige_Var (A : Variable) return String;
   function Zeige_Vars (A : Umgebung) return String;
   function Rev (T : Teile) return Teile;
   function Zeige_Konf (A : Konfiguration) return String;

-----------------------------------------------------------------------------------------------
---------- DELTA-Funktion (Zustandsbergangsfunktion) -----------------------------------------
--------------------------------------------------------------- teilweise eingerckt (***) ----
-----------------------------------------------------------------------------------------------
   function Delta2 (Z : Zustand) return Zustand;
   function Delta2_Ft (Z : Zustand; Ft : File_Type) return Zustand;
   
end Baumwandler_Package;
