WITH ada.strings.unbounded;
USE ada.strings.unbounded;
WITH text_io;
USE text_io;

PROCEDURE aufgabe4 IS 

   ------------------------------------------------------------------------------------------------
   -- Vorgaben                                                                                    --
   ------------------------------------------------------------------------------------------------

   TYPE detueber IS ARRAY (integer RANGE <>, character RANGE <>) OF integer; 
   -- ordnet einem Zustand und einem Zeichen
   -- einen Nachfolgezustand zu

   TYPE det_ueber_type 
         (n : positive) IS 
      RECORD 
         ueber : detueber (0 .. n, '!' .. '~');  
      END RECORD; 
   -- Einschraenkung auf Zustaende 0..n-1 und Fangzustand n
   -- (damit die Uebergangsfunktion total ist);
   -- T ist die Menge der druckbaren ASCII-Zeichen

   TYPE charfktn IS ARRAY (integer RANGE <>) OF boolean; 
   -- Charakteristische Funktion fuer Mengen
   -- ganzer Zahlen

   TYPE charfktn_type 
         (n : positive) IS 
      RECORD 
         inh : charfktn (0 .. n) := (OTHERS => false);  
      END RECORD; 
   -- beschraenkt die Obermenge auf die Zustaende
   -- 0 bis n;
   -- Initialisierung beschreibt die leere Menge

   TYPE dea_type 
         (n : positive) IS 
      RECORD 
         det : det_ueber_type (n);  
         f   : charfktn_type (n);  
      END RECORD; 
   -- beschreibt einen deterministischen endlichen
   -- Automaten mit
   --   Q = {0,1,...,n-1,n} (n ist Fangzustand)
   --   q0 = 0
   --   T = Menge der druckbaren ASCII-Zeichen
   --   delta = det
   --   F = F

   TYPE dea_ref IS ACCESS dea_type; 
   -- nuetzlich, wenn man z.B. mehrere DEAs in einem
   -- Array speichern moechte (siehe Scanner!)

   PROCEDURE pruefe (
         dea : dea_type; 
         w   : string    ) IS 
      akt : integer := 0;       -- aktueller Zustand                           
   BEGIN
      FOR i IN 1 .. w'last LOOP
         akt := dea.det.ueber (akt, w(i));
      END LOOP;
      put ("""" &w& """" & " gehoert");
      IF NOT dea.f.inh(akt) THEN
         put (" nicht");
      END IF;
      put (" zum Sprachschatz");
      -- new_line(2);
   END pruefe;
   ------------------------------------------------------------------------------------------------
   -- Ende Vorgaben                                                                              --
   ------------------------------------------------------------------------------------------------

   --Der Subtyp "druckbare" enthlt alle druckbaren ASCII-Zeichen:
   SUBTYPE druckbare IS character RANGE '!'..'~';

   --Werden falsche Eingaben gettigt, so wird ein Eingabefehler geworfen 
   --(siehe Funktion "scanner"):
   eingabefehler : EXCEPTION;  

   --Der Typ "tokenart" enthlt alle in der while-Sprache mglichen Token. Diese sind aufsteigend
   --nach Prioritt angeordnet:
   TYPE tokenart IS 
         (ganzezahl_token,  
          bezeichner_token, 
          rop_token,        
          op_token,         
          auf_token,        
          zu_token,         
          zuw_token,        
          strichp_token,    
          if_token,         
          then_token,       
          else_token,       
          fi_token,         
          skip_token,       
          while_token,      
          do_token,         
          od_token); 

   --Die Token-Liste ist folgendermaen deklariert:
   TYPE elem; 
   TYPE listoftoken IS ACCESS elem; 
   TYPE elem IS 
      RECORD 
         token  : tokenart;  
         string : unbounded_string;  
         rest   : listoftoken;  
      END RECORD; 

   ergebnis : listoftoken;  

   --Der Typ laengenarray wird in der Funktion "scanner" als bentigt
   TYPE laengenarray IS ARRAY (tokenart) OF integer; 
    
   --Mit dem Typ "scannerautomaten" wird jedem Token ein DEA zugeordnet:
   TYPE scannerautomaten IS ARRAY (tokenart) OF dea_ref; 


   ------------------------------------------------------------------------------------------------
   --4a)   Scandeas                                                                              --
   ------------------------------------------------------------------------------------------------
   --                                                                                            --
   --Jeder Token-Art wird in "scandeas" ein DEA zugewiesen. Hierzu werden die                    --
   --Komponente des DEAs definiert.                                                              --
   --                                                                                            --
   ------------------------------------------------------------------------------------------------
   --Beispielaufrufe:                                                                            --
   --                                                                                            --                                                                         --
   --Um die Beispiele zu verdeutlichen, wird die Funktion "pruefe" aus den Vorgaben genutzt.     --
   --                                                                                            --
   --Aufruf:   pruefe(scandeas(ganzezahl_token).ALL,"33");                                       --
   --Ausgabe:  "33" gehoert zum Sprachschatz                                                     --
   --                                                                                            --
   --Aufruf:   pruefe(scandeas(ganzezahl_token).ALL,"33a");                                      --
   --Ausgabe:  "33a" gehoert nicht zum Sprachschatz                                              --
   --                                                                                            --
   --Aufruf:   pruefe(scandeas(bezeichner_token).ALL,"then");                                    --
   --Ausgabe:  "then" gehoert zum Sprachschatz                                                   --
   --                                                                                            --
   --Aufruf:   pruefe(scandeas(then_token).ALL,"then");                                          --
   --Ausgabe:  "then" gehoert zum Sprachschatz                                                   --
   --                                                                                            --
   --Aufruf:   pruefe(scandeas(bezeichner_token).ALL,"thena");                                   --
   --Ausgabe:  "thena" gehoert zum Sprachschatz                                                  --
   --                                                                                            --
   --Aufruf:   pruefe(scandeas(bezeichner_token).ALL,"thena");                                   --
   --Ausgabe:  "thena" gehoert nicht zum Sprachschatz                                            --   
   ------------------------------------------------------------------------------------------------
   --                                                                                            --
   ------------------------------------------------------------------------------------------------

   --Am Beispiel des ganzezahl_token wird die Initialisierung erklrt:
    
   scandeas : scannerautomaten := (ganzezahl_token =>
   NEW dea_type'(n=>2,
      -- "n" ist Anzahl der Zustnde des Automaten 
      det=>det_ueber_type'(n=>2,
         ueber=>(0=>('0'..'9' => 1, OTHERS => 2),
            1=>('0'..'9' => 1, OTHERS => 2),
            2=>(OTHERS=>2))),
      -- "ueber" sind die bergangsfunktionen        
      f=>charfktn_type'(n=>2,inh=>(1=>true,OTHERS=>false))),
      -- "f" sind die gltigen finalen Zustnde

      bezeichner_token =>
   NEW dea_type'(n=>2,
      det=>det_ueber_type'(n=>2,
         ueber=>(0=>('a'..'z'|'A'..'Z' => 1, OTHERS => 2),
            1=>('0'..'9'|'a'..'z'|'A'..'Z' => 1, OTHERS => 2),
            2=>(OTHERS => 2))),
      f=>charfktn_type'(n=>2,inh=>(1=>true,OTHERS=>false))),

      rop_token =>
   NEW dea_type'(n=>4,
      det=>det_ueber_type'(n=>4,
         ueber=>(0=>('/' => 1, '<'|'>' => 2, '=' => 3, OTHERS => 4),
            1|2=>('=' => 3, OTHERS => 4),
            3|4=>(OTHERS=>4))),
      f=>charfktn_type'(n=>4,inh=>(2|3=>true,OTHERS=>false))),

      op_token =>
   NEW dea_type'(n=>6,
      det=>det_ueber_type'(n=>6,
         ueber=>(0=>('d' => 1, 'm' => 3, '+'|'*'|'-' => 5, OTHERS => 6),
            1=>('i' => 2, OTHERS => 6),
            2=>('v' => 5, OTHERS => 6),
            3=>('o' => 4, OTHERS => 6),
            4=>('d' => 5, OTHERS => 6),
            5|6=>(OTHERS => 6))),
      f=>charfktn_type'(n=>6,inh=>(5=>true,OTHERS=>false))),

      auf_token =>
   NEW dea_type'(n=>2,
      det=>det_ueber_type'(n=>2,
         ueber=>(0=>('(' => 1, OTHERS => 2),
            1|2=>(OTHERS => 2))),
      f=>charfktn_type'(n=>2,inh=>(1=>true,OTHERS=>false))),

      zu_token =>
   NEW dea_type'(n=>2,
      det=>det_ueber_type'(n=>2,
         ueber=>(0=>(')' => 1, OTHERS => 2),
            1|2=>(OTHERS => 2))),
      f=>charfktn_type'(n=>2,inh=>(1=>true,OTHERS=>false))),

      zuw_token =>
   NEW dea_type'(n=>3,
      det=>det_ueber_type'(n=>3,
         ueber=>(0=>(':' => 1, OTHERS => 3),
            1=>('=' => 2, OTHERS => 3),
            2|3=>(OTHERS => 3))),
      f=>charfktn_type'(n=>3,inh=>(2=>true,OTHERS=>false))),

      strichp_token =>
   NEW dea_type'(n=>2,
      det=>det_ueber_type'(n=>2,
         ueber=>(0=>(';' => 1, OTHERS => 2),
            1|2=>(OTHERS => 2))),
      f=>charfktn_type'(n=>2,inh=>(1=>true,OTHERS=>false))),

      if_token =>
   NEW dea_type'(n=>3,
      det=>det_ueber_type'(n=>3,
         ueber=>(0=>('i' => 1, OTHERS => 3),
            1=>('f' => 2, OTHERS => 3),
            2|3=>(OTHERS => 3))),
      f=>charfktn_type'(n=>3,inh=>(2=>true,OTHERS=>false))),

      then_token =>
   NEW dea_type'(n=>5,
      det=>det_ueber_type'(n=>5,
         ueber=>(0=>('t' => 1, OTHERS => 5),
            1=>('h' => 2, OTHERS => 5),
            2=>('e' => 3, OTHERS => 5),
            3=>('n' => 4, OTHERS => 5),
            4|5=>(OTHERS => 5))),
      f=>charfktn_type'(n=>5,inh=>(4=>true,OTHERS=>false))),

      else_token =>
   NEW dea_type'(n=>5,
      det=>det_ueber_type'(n=>5,
         ueber=>(0=>('e' => 1, OTHERS => 5),
            1=>('l' => 2, OTHERS => 5),
            2=>('s' => 3, OTHERS => 5),
            3=>('e' => 4, OTHERS => 5),
            4|5=>(OTHERS => 5))),
      f=>charfktn_type'(n=>5,inh=>(4=>true,OTHERS=>false))),

      fi_token =>
   NEW dea_type'(n=>3,
      det=>det_ueber_type'(n=>3,
         ueber=>(0=>('f' => 1, OTHERS => 3),
            1=>('i' => 2, OTHERS => 3),
            2|3=>(OTHERS => 3))),
      f=>charfktn_type'(n=>3,inh=>(2=>true,OTHERS=>false))),

      skip_token =>
   NEW dea_type'(n=>5,
      det=>det_ueber_type'(n=>5,
         ueber=>(0=>('s' => 1, OTHERS => 5),
            1=>('k' => 2, OTHERS => 5),
            2=>('i' => 3, OTHERS => 5),
            3=>('p' => 4, OTHERS => 5),
            4|5=>(OTHERS => 5))),
      f=>charfktn_type'(n=>5,inh=>(4=>true,OTHERS=>false))),

      while_token =>
   NEW dea_type'(n=>6,
      det=>det_ueber_type'(n=>6,
         ueber=>(0=>('w' => 1, OTHERS => 6),
            1=>('h' => 2, OTHERS => 6),
            2=>('i' => 3, OTHERS => 6),
            3=>('l' => 4, OTHERS => 6),
            4=>('e' => 5, OTHERS => 6),
            5|6=>(OTHERS => 6))),
      f=>charfktn_type'(n=>6,inh=>(5=>true,OTHERS=>false))),

      do_token =>
   NEW dea_type'(n=>3,
      det=>det_ueber_type'(n=>3,
         ueber=>(0=>('d' => 1, OTHERS => 3),
            1=>('o' => 2, OTHERS => 3),
            2|3=>(OTHERS => 3))),
      f=>charfktn_type'(n=>3,inh=>(2=>true,OTHERS=>false))),

      od_token =>
   NEW dea_type'(n=>3,
      det=>det_ueber_type'(n=>3,
         ueber=>(0=>('o' => 1, OTHERS => 3),
            1=>('d' => 2, OTHERS => 3),
            2|3=>(OTHERS => 3))),
      f=>charfktn_type'(n=>3,inh=>(2=>true,OTHERS=>false)))
      );




   ------------------------------------------------------------------------------------------------
   --4b) erkanntesprafix                                                                         --
   ------------------------------------------------------------------------------------------------
   --                                                                                            --
   --Die Funktion "erkanntespraefix" prft ob der Anfang eines gegebenen Strings zum Sprachschatz-- 
   --des mitgegebnen DEAs gehrt und gibt die Lnge des passenden Anfangs zurck.                --
   --Dazu wird der String Zeichen fr Zeichen geprft.                                           --
   --Es gibt 2 Mglichkeiten, dass die Prfung abgebrochen wird.                                 --
   --Mglichkeit1: Die Funktion ist beim Prfen am Ende des Strings angelangt.                   --
   --              Ist der aktuelle Zustand des DEAs ein finaler Zustand, so wird die bis dahin  --
   --              ermittelte Lnge zurckgegeben.                                               --
   --Mglichkeit2: Der DEA befindet sich im "Fangzustand".                                       --
   --              Sofern der Zustand vor Eintritt in den Fangzustand ein finaler Zustand war,   --
   --              wird die bis dahin ermittelte Lnge zurckgegeben.                            --
   --Liegt kein finaler Zustand vor, so wird 0 zurckgegeben, da der String nicht mit einem      --
   --gltigen Prfix beginnt.                                                                    --
   --                                                                                            --
   ------------------------------------------------------------------------------------------------
   --                                                                                            --
   --Beispielaufrufe:                                                                            --
   --                                                                                            --
   --Aufruf:    erkanntespraefix(scandeas(ganzezahl_token),"33");                                --
   --Ergebnis:  2                                                                                --
   --Aufruf:    erkanntespraefix(scandeas(ganzezahl_token),"33a");                               --
   --Ergebnis:  2                                                                                --
   --Aufruf:    erkanntespraefix(scandeas(ganzezahl_token),"a33");                               --
   --Ergebnis:  0                                                                                --
   --Aufruf:    erkanntespraefix(scandeas(bezeichner_token),"then");                             --
   --Ergebnis:  4                                                                                --
   --Aufruf:    erkanntespraefix(scandeas(then_token),"then");                                   --
   --Ergebnis:  4                                                                                --
   --Aufruf:    erkanntespraefix(scandeas(then_token),"thena");                                  --
   --Ergebnis:  4                                                                                --
   --                                                                                            --
   ------------------------------------------------------------------------------------------------
   --                                                                                            --
   ------------------------------------------------------------------------------------------------

   FUNCTION erkanntespraefix (
         d : dea_ref; 
         w : string   ) 
     RETURN natural IS 
      akt,  
      aktvorher : integer := 0;  
      ausgabe   : natural := 0;  
      i         : integer := 1;  
   BEGIN

      WHILE akt /= d.n AND THEN i /= w'last+1 LOOP

         IF w(i) NOT IN druckbare THEN
            aktvorher:=akt;
            akt:= d.n;
            ausgabe:= ausgabe + 1;
         ELSE
            aktvorher:=akt;
            akt := d.det.ueber (akt, w(i));
            i:=i+1;
            ausgabe:= ausgabe + 1;
         END IF;
      END LOOP;

      IF d.f.inh(akt) THEN
         RETURN ausgabe;
      ELSIF d.f.inh(aktvorher) THEN
         RETURN ausgabe -1;
      ELSE
         RETURN 0;
      END IF;
   END erkanntespraefix;



   ------------------------------------------------------------------------------------------------
   --4c) Scanner                                                                                 --                                                                                          --
   ------------------------------------------------------------------------------------------------
   --                                                                                            --
   --Die Funktion "scanner" erstellt aus einem gegebenen String rekursiv eine Token-Liste.       --
   --Dazu werden alle whitespaces (Leerzeichen, Tabs, Zeilenumbrche) am Anfang des Strings      --
   --berlesen und fr die weitere Bearbeitung "abgeschnitten".                                  --
   --Besteht der String nur aus whitespaces oder ist er leer, so wird eine leere                 --
   --Token-Liste zurckgegeben.                                                                  --
   --Mit der Funktion "erkanntespraefix" wird jetzt fr jede Token-Art geprft, ob der String mit--
   --einem gltigen Wort beginnt und wie lang dieses ist. Die Lnge wird gespeichert.            --
   --Nun wird herausgefunden, fr welche Token-Art das lngste Prfix gefunden wurde. Sind zwei  --
   --Prfixe von gleicher Lnge, so "gewinnt" die Token-Art mit hherer Prioritt.               --
   --Wird kein Prfix gefunden, ist der eingegebene String fehlerhaft und der                    --
   --Eingabefehler wird ausgelst.                                                               --
   --Das lngste, gefundene Prfix wird jetzt vom String "abgeschnitten" und ein Element der     --
   --Token-Liste erstellt, welches die Token-Art und das gefundene Prfix als String enthlt.    --
   --Der restliche String wird rekursiv mit der Funktion "scanner" weiter abgearbeitet.          --
   --Sobald ein Zeichen als falsche Eingabe identifiziert wird, wird dieses ausgegeben und der   --
   --Eingabefehler wird geworfen.                                                                --
   --                                                                                            --
   ------------------------------------------------------------------------------------------------
   --                                                                                            --
   --Beispielaufrufe:                                                                            --
   --                                                                                            --
   --Aufruf:   ergebnis:= scanner(to_unbounded_string("if a=[-2]"));                             --
   --          druckelistoftoken(ergebnis);                                                      --
   --                                                                                            --
   --Ausgabe:  EINGABEFEHLER                                                                     --
   --          Das Zeichen '[' ist nicht im Sprachschatz der Token                               --
   --          raised AUFGABE4.EINGABEFEHLER : aufgabe4.adb:465                                  --
   --                                                                                            --
   --Aufruf:   ergebnis:= scanner(to_unbounded_string("if a= -2 do"&character'val(27)&"od"));    --
   --          druckelistoftoken(ergebnis);                                                      --
   --                                           --character'val(27) entspricht der ESC-Anweisung --
   --                                                                                            --                   
   --Ausgabe:  EINGABEFEHLER                                                                     --
   --          Das ASCII-Zeichen Nr. 27 ist kein druckbares Zeichen                              --
   --          raised AUFGABE4.EINGABEFEHLER : aufgabe4.adb:444                                  --
   --                                                                                            --
   --weitere Beispielaufrufe folgen in Aufgabe d)                                                --                                         --
   --                                                                                            --
   ------------------------------------------------------------------------------------------------
   --                                                                                            --
   ------------------------------------------------------------------------------------------------

   FUNCTION scanner (
         w : unbounded_string ) 
     RETURN listoftoken IS 
      laengepraefix   : laengenarray     := (OTHERS => 0);  
      laengste        : integer          := 0;  
      gefundenertoken : tokenart;  
      zwischenstring,  
      ergebnisstring,  
      reststring      : unbounded_string;  
      position        : integer          := 1;  

   BEGIN
      WHILE position <= length (w) AND THEN (
            element (w,position) = character'val(32) OR    --Leerzeichen
            element (w,position) = character'val(9) OR     --Tabulator
            element (w,position) = character'val(13) OR    --Carriage-Return \ entspricht
            element (w,position) = character'val(10)) LOOP --Line Feed       / Zeilenumbruch
         position:= position +1;
      END LOOP;

      IF position > length (w) THEN
         RETURN NULL;
      ELSIF element (w,position) NOT IN druckbare THEN
         put_line("Das ASCII-Zeichen Nr." & integer'image(character'pos(
                  element (w,position))) & " ist kein druckbares Zeichen");
         RAISE eingabefehler;
      END IF;

      zwischenstring := delete(w,1,position-1);

      FOR i IN tokenart LOOP
         laengepraefix(i):= erkanntespraefix(scandeas(i),to_string(
               zwischenstring));
      END LOOP;

      FOR i IN tokenart LOOP
         IF laengepraefix(i) >= laengste THEN
            laengste:=laengepraefix(i);
            gefundenertoken:=i;
         END IF;
      END LOOP;

      IF laengste = 0 THEN
         put_line("EINGABEFEHLER:");
         put_line("Das Zeichen '" &element (zwischenstring,1)&
            "' ist nicht im Sprachschatz der Token");
         RAISE eingabefehler;
      ELSIF length(zwischenstring) = laengste THEN
         ergebnisstring:=zwischenstring;
      ELSE
         ergebnisstring:= delete (zwischenstring,laengste+1,length(
               zwischenstring));
      END IF;

      reststring:= delete (zwischenstring,1,laengste);

      RETURN NEW elem'(gefundenertoken,ergebnisstring,scanner(reststring));

   END scanner;



   ------------------------------------------------------------------------------------------------
   --4d) Anwendung                                                                               --
   ------------------------------------------------------------------------------------------------
   --                                                                                            --
   --Im Folgenden werden 3 while-Programmstcken initialisiert.                                  --
   --Diese Strings werden im Anweisungsteil aufgerufen                                           --
   --                                                                                            --
   ------------------------------------------------------------------------------------------------
   --                                                                                            --
   ------------------------------------------------------------------------------------------------
   
   --Beispiel 1:                                                                                 --
   --Programmstck:     while b <= x do                                                          --
   --                           a := (a + 1);                                                    --
   --                           c := (c + 2);                                                    --
   --                           b := (b + c);                                                    --
   --                   od                                                                       --
                                                                                           
   while1 : string := ("while b <= x do" & character'val (13) & character'val                    
     (10) & character'val (9) & "a := (a + 1);" & character'val (13) & character'val            
     (10) & character'val (9) & "c := (c + 2);" & character'val (13) & character'val
     (10) & character'val (9) & "b := (b + c);" & character'val (13) & character'val
     (10) & "od");  
                                                                                            
   
   --Beispiel 2:                                                                                 --
   --Programmstck:     while i /= 1 do                                                          --
   --                           f := i * i;                                                      --
   --                           i := i - 1;                                                      --
   --                   od                                                                       --   
                                                                                  
   while2 : string := ("while i /= 1 do" & character'val (13) & character'val
     (10) & character'val (9) & "f := f * i;" & character'val (13) & character'val
     (10) & character'val (9) & "i := i - 1;" & character'val (13) & character'val
     (10) & "od");  


   --Beispiel 3:                                                                                 --
   --Programmstck:     while j /= n do                                                          --
   --                           f := f div (n - j);                                              --
   --                           j := j + 1;                                                      --
   --                   od                                                                       -- 
   
   while3 : string := ("while j /= n do" & character'val (13) & character'val
     (10) & character'val (9) & "f := f div (n - j);" & character'val (13)
     & character'val (10) & character'val (9) & "j := j + 1;" & character'val
     (13) & character'val (10) & "od");  
   
   ------------------------------------------------------------------------------------------------
   -- druckelistoftoken (Zusatz)
   --
   -- Die Funktion "druckelistoftoken" dient der Veranschaulichung der Prozedur:
   ------------------------------------------------------------------------------------------------
    
   PROCEDURE druckelistoftoken (
         lot : listoftoken ) IS 
      lotkopie : listoftoken := lot;  
   BEGIN
      new_line;
      WHILE lotkopie /= NULL LOOP
         put("(");
         put(tokenart'image(lotkopie.token));
         put("," & """");
         put(to_string(lotkopie.string));
         put("""" & ")");
         new_line;
         lotkopie:=lotkopie.rest;
      END LOOP;

   END druckelistoftoken;

BEGIN

   --Die Aufrufe der 3 While-Programmstcke aus Aufgabe 4d)
   put_line("While-Programmstueck 1:");
   put_line(while1);
   ergebnis:= scanner(to_unbounded_string(while1));
   druckelistoftoken(ergebnis);
   new_line;
   
   --Ausgabe:     While-Programmstueck 1:
   --             while b <= x do
   --                      a := (a + 1);
   --                      c := (c + 2);
   --                      b := (b + c);  
   --             od
   --
   --             (WHILE_TOKEN,"while") 
   --             (BEZEICHNER_TOKEN,"b") 
   --             (ROP_TOKEN,"<=") 
   --             (BEZEICHNER_TOKEN,"x") 
   --             (DO_TOKEN,"do") 
   --             (BEZEICHNER_TOKEN,"a") 
   --             (ZUW_TOKEN,":=") 
   --             (AUF_TOKEN,"(") 
   --             (BEZEICHNER_TOKEN,"a") 
   --             (OP_TOKEN,"+") 
   --             (GANZEZAHL_TOKEN,"1") 
   --             (ZU_TOKEN,")") 
   --             (STRICHP_TOKEN,";") 
   --             (BEZEICHNER_TOKEN,"c") 
   --             (ZUW_TOKEN,":=") 
   --             (AUF_TOKEN,"(") 
   --             (BEZEICHNER_TOKEN,"c") 
   --             (OP_TOKEN,"+") 
   --             (GANZEZAHL_TOKEN,"2") 
   --             (ZU_TOKEN,")") 
   --             (STRICHP_TOKEN,";") 
   --             (BEZEICHNER_TOKEN,"b") 
   --             (ZUW_TOKEN,":=") 
   --             (AUF_TOKEN,"(") 
   --             (BEZEICHNER_TOKEN,"b") 
   --             (OP_TOKEN,"+") 
   --             (BEZEICHER_TOKEN,"c") 
   --             (ZU_TOKEN,")") 
   --             (STRICHP_TOKEN,";")
   --             (OD_TOKEN,"od") 


   put_line("While-Programmstueck 2:");
   put_line(while2);
   ergebnis:= scanner(to_unbounded_string(while2));
   druckelistoftoken(ergebnis);
   new_line;
   
   --Ausgabe:     While-Programmstueck 2:
   --             while i /= 1 do
   --                      f := i * i;
   --                      i := i - 1;  
   --             od
   --
   --             (WHILE_TOKEN,"while") 
   --             (BEZEICHNER_TOKEN,"i") 
   --             (ROP_TOKEN,"/=") 
   --             (GANZEZAHL_TOKEN,"1") 
   --             (DO_TOKEN,"do") 
   --             (BEZEICHNER_TOKEN,"f") 
   --             (ZUW_TOKEN,":=")  
   --             (BEZEICHNER_TOKEN,"f") 
   --             (OP_TOKEN,"*") 
   --             (BEZEICHNER_TOKEN,"i") 
   --             (STRICHP_TOKEN,";") 
   --             (BEZEICHNER_TOKEN,"i") 
   --             (ZUW_TOKEN,":=") 
   --             (BEZEICHNER_TOKEN,"i") 
   --             (OP_TOKEN,"-") 
   --             (GANZEZAHL_TOKEN,"1")
   --             (STRICHP_TOKEN,";")  
   --             (OD_TOKEN,"od")


   put_line("While-Programmstueck 3:");
   put_line(while3);
   ergebnis:= scanner(to_unbounded_string(while3));
   druckelistoftoken(ergebnis);

   --Ausgabe:     While-Programmstueck 3:
   --             while j /= n do
   --                      f := f div (n - j);
   --                      j := j + 1;  
   --             od
   --
   --             (WHILE_TOKEN,"while") 
   --             (BEZEICHNER_TOKEN,"j") 
   --             (ROP_TOKEN,"/=") 
   --             (BEZEICHNER_TOKEN,"n") 
   --             (DO_TOKEN,"do") 
   --             (BEZEICHNER_TOKEN,"f") 
   --             (ZUW_TOKEN,":=")  
   --             (BEZEICHNER_TOKEN,"f") 
   --             (OP_TOKEN,"div")
   --             (AUF_TOKEN,"(") 
   --             (BEZEICHNER_TOKEN,"n")
   --             (OP_TOKEN,"-")
   --             (BEZEICHNER_TOKEN,"j")
   --             (ZU_TOKEN,")")
   --             (STRICHP_TOKEN,";") 
   --             (BEZEICHNER_TOKEN,"j") 
   --             (ZUW_TOKEN,":=") 
   --             (BEZEICHNER_TOKEN,"j") 
   --             (OP_TOKEN,"+") 
   --             (GANZEZAHL_TOKEN,"1")
   --             (STRICHP_TOKEN,";")  
   --             (OD_TOKEN,"od")

END aufgabe4;
