generic

   type Zeichen is (<>);
   -- Aufzaehlungstyp mit allen Grammatiksymbolen
   -- Zuerst kommt das Startsymbol, 
   -- dann ggfs weitere Nichtterminale
   -- danach die Terminale
   
   AnzahlNichtterminale : in positive;
   -- die Nichtterminale haben die Indexpositionen 0 .. AnzahlNichtterminale-1,
   -- die Terminale die Indexpositionen ab AnzahlNichtterminale
   
package GrammatikGenerator is
 
   -- aus der Beschreibung der generischen Parameter folgen die Definitionen:
   subtype Nichtterminal is 
      Zeichen range Zeichen'first .. Zeichen'val(AnzahlNichtterminale-1);
   subtype Terminal is 
      Zeichen range Zeichen'val(AnzahlNichtterminale) .. Zeichen'last;
   Startsymbol : constant Nichtterminal := Zeichen'first;
   
   function istTerminal (z : Zeichen) return Boolean;
   -- hilft, zwischen Terminalen und Nichtterminalen zu unterscheiden
   
   type Zeichenfolge is private;
   -- eine Folge von Grammatikzeichen, 
   -- z.B. die rechte Seite einer Regel
   -- oder eine Satzform waehrend einer Ableitung
   
   type Zeichenarray is array (integer range <>) of Zeichen;
   -- vereinfacht Definition rechter Regelseiten
   
   type Wort is array (integer range <>) of Terminal;
   -- verwendet zur Definition terminaler Zeichenfolgen
   
   type Alternativen is private;
   -- eine Folge von Alternativen (rechten Regelseiten) zum gleichen Nichtterminal
   
   type Regelmenge is array (Nichtterminal) of Alternativen;
   -- ordnet jedem Nichtterminal die Folge seiner Alternativen zu;
   -- anfangs hat jedes Nichtterminal eine leere Folge von Alternativen
   
   function hatEpsilonRegel (n : Nichtterminal) return Boolean;
   -- hilft festzustellen, ob eine der Alternativen zu n eine
   -- leere rechte Seite (dargestellt als leere Zeichenfolge) hat
   
   procedure neueAlternativeHinzu(links : Nichtterminal; rechts : Zeichenarray);
   -- fuegt der Regelmenge eine neue Alternative von links hinzu;
   -- es wird nicht geprueft, ob die Alternative tatsaechlich neu ist
   
   procedure druckeGrammatik;
   -- druckt die Grammatik wie im Beispiel auf dem Aufgabenblatt
   
   function istGrammatikEinfach return Boolean;
   -- eine Grammatik heisst einfach, wenn fuer jedes Nichtterminal
      -- die nichtleeren Alternativen alle mit Terminalzeichen beginnen und
      -- diese Terminalzeichen paarweise verschieden sind
   
   function wortImSprachschatz (w : Wort) return Boolean;
   -- zu einer einfachen Grammatik G und einem gegebenen Wort w
   -- feststellen, ob w im Sprachschatz L(G) liegt
   
private

   type ZeichenfolgeRec;
   type Zeichenfolge is access ZeichenfolgeRec;
   type ZeichenfolgeRec is record 
      erstes  : Zeichen;
      weitere : Zeichenfolge;
   end record;

   type AlternativenRec;
   type Alternativen is access AlternativenRec;
   type AlternativenRec is record 
      erste  : Zeichenfolge;
      andere : Alternativen;
   end record;
   
end GrammatikGenerator;