-- generisches Package zum bearbeiten von Adjazenzmatrizen
-- Package Body

-- Testaufrufe zu diesen Funktionen finden Sie in der Instantiierungen
-- fr den Typ Integer (instantiierungint.adb) und Character 
-- (instantiierungchr.adb) 

-- Aufruf bentigter Packages fr die Funktion Drucke
with Text_Io;
use Text_Io;

package body Matrix is

-- Die Funktion bildet aus zwei Matrizen das Matrizenprodukt. Hierzu 
--  erhlt die Funktion zwei Matrizen und gibt eine neue Matrix zurck
--  Die Matrix 1 wird hierbei zeilenweise, die Matrix 2 spaltenweise 
--  durchlaufen. Es werden jeweils die gleichen Elemente der beiden 
--  Matrizen mit "AND" verknpft. Die "OR" Verknpfung stellt die 
--  Aufsummierung der jeweiligen Spaltenelemente dar. Die Rckgabe 
--  entpricht dem mathematischen Matrizenprodukt der beiden Matrizen.
   
   function Produkt (
         M1,           
         M2 : Z_Matrix ) 
     return Z_Matrix is 

      C : Z_Matrix;  

   begin
      for I in Z'range loop
         for J in Z'range loop
            C(I,J) := False;
            for K in Z'range loop
               C(I,J) := C(I,J) or (M1(I,K) and M2(K,J));
            end loop;
         end loop;
      end loop;
      return C;
   end Produkt;


-- Die Funktion bildet die Summe aus zwei Matrizen. Hierzu 
--  erhlt die Funktion 2 Matrizen und gibt eine neue 
--  Matrix zurck. Es werden hierbei jeweils die 
--  Wahrheitswerte mit der gleichen Position in zwei Matrizen 
--  mit "OR" verknpft. Hierzu werden die Matrizen erst spalten-, 
--  dann zeilenweise durchlaufen. Die Rckgabe der Funktion 
--  entspricht der mathematischen Summe zweier Matrizen.
   
   function Summe (
         M1,           
         M2 : Z_Matrix ) 
     return Z_Matrix is 

      C : Z_Matrix;  

   begin
      for I in Z'range loop
         for J in Z'range loop
            C(I,J) := M1(I,J) or M2(I,J);
         end loop;
      end loop;
      return C;
   end Summe ;


-- Die Funktion transponiert die Matrix, hierzu erhlt die Funktion 
--  eine Matrix und gibt eine neue Matrix zurck.
--  Die Funktion durchluft die Matrix zeilenweise, wobei der 
--  Spaltenindex nach jedem Zeilendurchlauf ebenfalls um eins 
--  erhht wird. Da das Element der Hauptdiagonalen nicht gespiegelt 
--  werden muss, wird der Spaltenindex der Zeile und der Zeilenindex 
--  der Spalte um eins erhht (Variable J). Der Vertauschungsvorgang 
--  wird mit Hilfe einer Zuweisung des Zeilenelements auf die 
--  Variable H durchgefhrt. Anschlieend wird das Zeilenelement 
--  durch das Spaltenelement ersetzt. Im letzten Schritt wird die 
--  Variable H an das Spaltenelement bergeben. Die Rckgabe 
--  der Funktion entspricht der mathematischen Transposition 
--  einer Matrix was der Spiegelung an der 
--  Hauptdiagonalen entspricht.
 
   function Transponiere (
         M1 : Z_Matrix ) 
     return Z_Matrix is 

      C : Z_Matrix := M1;  
      H : Boolean;  
      J : Z;  

   begin
      for I in Z'range loop
         if I=Z'Last then
            exit;
         end if;
         J:=Z'Succ(I);
         loop
            H := C(I,J);
            C(I,J) := C(J,I);
            C(J,I) := H;
            exit when J=Z'Last;
            J:=Z'Succ(J);
         end loop;
      end loop;
      return C;
   end Transponiere;


-- Die Funktion berechnet die Transitive Huelle einer Matrix, 
--  sie erhlt hierzu eine Matrix und gibt eine neue 
--  Matrix zurck. Hierbei sttzt sich die Funktion auf die 
--  Funktion Summe und die Funktion Produkt ab, da die 
--  Verknpfung und Hintereinanderausfhrung der 
--  mathematischen Transitivitt entspricht. Die Schleife 
--  wird hierbei so oft durchlaufen wie Elemente im Typ Z 
--  enthalten sind. Die Rckgabe der Funktion entspricht 
--  der mathematischen transitiven Huelle einer Matrix.
    
   function Transhuelle (
         M1 : Z_Matrix ) 
     return Z_Matrix is 

      C : Z_Matrix := M1;  
   begin
      for I in Z'range loop
         C := Summe (M1,(Produkt(C,M1)));
      end loop;
      return C;
   end Transhuelle;


-- Die Funktion berechnet die reflexiv-transitive Huelle 
--  einer Matrix, sie erhlt hierzu  eine Matrix und gibt 
--  eine Matrix zurck. Die Funktion sttzt sich auf die 
--  Funktion TransHuelle ab. Die Matrix wird gleichzeitig 
--  zeilen- und spaltenweise durchlaufen. Um die reflexive 
--  Eigenschaft zu erreichen werden jeweils Zeile und 
--  Spalte miteinander verglichen, befindet sich 
--  mindestens in einem von beiden der Wahrheitswert "TRUE",
--  wird der Schnittpunkt der beiden Achsen auch 
--  auf "TRUE" gesetzt. Die Rckgabe der Funktion 
--  entspricht der mathematischen reflexiv-transitiven 
--  Huelle einer Matrix.
 
   function Reflextranshuelle (
         M1 : Z_Matrix ) 
     return Z_Matrix is 

      C : Z_Matrix := M1;  

   begin
      C:= Transhuelle(M1);
      for I in Z'range loop
               C(I,I):=True;
      end loop;
      return C;
   end;


   -- Diese Prozedur gibt die bergebene Matrix auf dem Bildschirm aus.
   -- Hierzu wird die Matrix und ein Bezeichner bergeben.
   -- Zuerst wird der Bezeichner ausgegeben anschlieend die Matrix.
   -- Die Funktion unterscheidet zwischen Integer und anderen Typen,
   -- da hierbei der Spaltenabstand Matrixkopfzeile variiert.
   -- Die Wahrheitswerte der Matrix werden hier mit "0" fr "FALSE" 
   -- und "1" fr "TRUE" ausgegeben.

   procedure Drucke (
         M1         : Z_Matrix; 
         Bezeichner : String    ) is 

      Leerzeichen : Boolean := False;  

   begin
      New_Line(2);
      Put(Bezeichner);
      New_Line(2);

      if Z'Image(Z'First)=Integer'Image(1) then
         Leerzeichen:=True;
      end if;

      if Leerzeichen=False then
         Put(" ");
      end if;

      Put("     ");
      for I in Z'range loop
         Put(Z'Image(I));
         if Leerzeichen then
            Put(" ");
         end if;
      end loop;
      New_Line(2);
      for I in Z'range loop
         Put(Z'Image(I));
         Put(" :");
         for J in Z'range loop
            if M1(I,J) then
               Put("  1");
            else
               Put("  0");
            end if;
         end loop;
         New_Line;
      end loop;
      New_Line;
   end Drucke;


   --   Die Funktion berechnet die Transition von zwei Matrizen, 
   --   sie erhlt hierzu zwei Matrizen
   --   und gibt eine neue Matrix zurck.
   --   Die Matrix 1 wird so lange zeilenweise durchlaufen, 
   --   bis der Wahrheitswert "TRUE" gefunden
   --   wird oder die Matrix komplett durchlaufen wurde.
   --   Im Fall, dass der Wahrheitswert "TRUE" gefunden wurde, 
   --   wird die zweite Matrix in der Zeile, entsprechend dem 
   --   Spaltenindex der ersten Matrix, durchlaufen.
   --   Dies geschieht zeilenweise so lange, bis ebenfalls 
   --   der Wahrheitswert "TRUE" gefunden wird oder das Ende 
   --   der Zeile erreicht wird. Wird in der jeweiligen Zeile 
   --   der zweiten Matrix ebenfalls der Wahrheitswert "TRUE" 
   --   gefunden, so wird die entsprechende Position in 
   --   der neuen Matrix auf "TRUE" gesetzt. Die Position 
   --   entspricht hierbei dem Zeilenindex der ersten Matrix 
   --   und dem Spaltenindex der zweiten Matrix.

   function Transition (
         Z1,           
         Z2 : Z_Matrix ) 
     return Z_Matrix is 
      M3 : Z_Matrix := (others => (others => False));  
   begin

      for I in Z'range loop
         for J in Z'range loop
            if Z1(I,J) then
               for K in Z'range loop
                  if Z2(J,K) then
                     M3(I,K):=True;
                  end if;
               end loop;
            end if;
         end loop;
      end loop;
      return M3;
   end Transition;

end Matrix;