-- Generisches Package das Funktionen auf Mengen bereitstellt
-- Package Body

package body Mengen is

   -- Festlegung der Leeren Menge.
   --  Alle Elemente werden hierzu mit "FALSE" belegt.
   LeereMenge : constant ListZ := (others => False);  


   -- Erzeugt eine neue leere Menge und
   --  gibt eine neue leere Liste zurck.
   function NeueLeerMenge return ListZ is 
   begin
      return LeereMenge;
   end NeueLeerMenge;


   -- Fgt ein neues Element in eine Liste ein.
   --  Hierzu werden ein Element und eine Liste bergeben
   --  und die genderte Liste zurckgegeben.
   --  Dazu wird der Wahrheitswert an der betreffenden 
   --  Stelle auf "TRUE" gesetzt.
   procedure Einfgen (E : Z; Ch : in out Listz) is 
   begin
      Ch(E) := True;
   end Einfgen;


   -- Lscht ein Element aus einer Liste.
   --  Hierzu werden ein Element und eine Liste bergeben
   --  und die genderte Liste zurckgegeben.
   --  Dazu wird der Wahrheitswert an der betreffenden
   --  Stelle auf "FALSE" gesetzt.
   procedure Loesche (E : Z; Ch : in out ListZ) is 
   begin
      Ch(E) := False;
   end Loesche;


   -- Prft, ob ein bestimmtes Element in einer Liste 
   --  vorhanden ist.
   --  Hierzu werden ein Element und eine Liste bergeben
   --  und ein Wahrheitswert zurckgegeben. Ist das Element
   --  enthalten, wird der Wert "TRUE", andernfalls wird 
   --  "FALSE" zurckgegeben.
   function Vorhanden (E : Z; Ch : ListZ) return Boolean is 
   begin
      return Ch(E);
   end Vorhanden;


   -- Prft, ob eine Liste leer ist.
   --  Hierzu wird eine Liste bergeben
   --  und ein Wahrheitswert zurckgegeben.
   --  Die Liste wird durchlaufen.
   --  Sofern ein Element enthalten ist, bricht die Schleife ab
   --  und es wird "FALSE" zurckgegeben.
   --  Ansonsten wird die Liste bis zum Ende durchlaufen und 
   --  der Wert "TRUE" zurckgegeben.
   function Istleer (Ch : Listz) return Boolean is 
   begin
      for I in Z loop
         if Ch(I) then
            return False;
         end if;
      end loop;
      return True;
   end IstLeer;


   -- Bildet die Vereinigungsmenge zweier Listen.
   --  Hierzu werden zwei Listen bergeben
   --  und eine neue Liste (Vereinigung) zurckgegeben. 
   --  Es werden alle Elemente beider Listen durchlaufen 
   --  und mit "OR" verknpft. 
   --  D.h. hat an der betreffenden Indexposition 
   --  mindestens  eines der Elemente den Wert "TRUE", 
   --  wird in der neuen Liste (Vereinigung) der Wert an 
   --  dieser Position mit "TRUE" belegt.
   function Vereinigung (Ch1, Ch2 : ListZ) return ListZ is 
      H : ListZ;  
   begin
      for I in Z loop
         H(I) := Ch1(I) or Ch2(I);
      end loop;
      return H;
   end Vereinigung;


   -- Bildet die Schnittmenge zweier Listen.
   --  Hierzu werden zwei Listen bergeben
   --  und eine neue Liste (Schnitt) zurckgegeben. 
   --  Es werden alle Elemente beider Listen durchlaufen 
   --  und mit "AND" verknpft.
   --  D.h. haben an der betreffenden Indexposition 
   --  beide Elemente den Wert "TRUE", wird in der neuen
   --  Liste (Schnitt) der Wert an dieser Position mit
   --  "TRUE" belegt.
   function Schnitt (Ch1, Ch2 : ListZ) return ListZ is 
      H : Listz;  
   begin
      for I in Z loop
         H(I) := Ch1(I) and Ch2(I);
      end loop;
      return H;
   end Schnitt;

end Mengen;