-- generisches Package zur Erzeugung einer Adjazenzmatrix 
-- aus einer Liste
-- Package Body
with Text_Io;
use Text_Io;

package body Toadi is


   -- Die Funktion erzeugt eine Adjazenzmatrix.
   --  Hierzu erhlt sie einen Pointer auf eine Tupelliste
   --  und gibt eine Matrix zurck.
   --  Es wird eine Matrix erzeugt und mit "FALSE" belegt.
   --  Falls der bergebene Zeiger kein Ziel hat ("null"),
   --  wird eine leere Matrix zurckgegeben.
   --  Andernfalls wird jedes Element der Liste durchlaufen
   --  und der jeweilige Wahrheitswert der Matrix auf "TRUE" gesetzt.
   function Erzeugeadimatrix (Zeiger : Tupelpoint) return Z_Matrix is
      Adimatrix : Z_Matrix   := (others => (others => False));  
      Point     : Tupelpoint := Zeiger;  
   begin
      if Zeiger = null then
         return Adimatrix;
      end if;
      loop
         Adimatrix(Point.Pos1,Point.Pos2):=True;
         Point:=Point.Rest;
         exit when Point = null;
      end loop;
      return Adimatrix;
   end Erzeugeadimatrix;



   -- Die Funktion erzeugt aus einer Adjazenzmatix eine Menge aus  
   --  Tupeln. Hierzu erhlt die Funktion eine Matrix und gibt einen
   --  Zeiger auf das erste Element der Tupelliste zurck. Die Matrix 
   --  wird vom letzten bis zum ersten Element zeilenweise durchlaufen.
   --  Falls der Wert des betrachteten Elements "TRUE" ist, wird ein 
   --  neues Tupel erzeugt und der Zeiger auf dieses Element gesetzt.
   function Erzeugemenge (Matrix : Z_Matrix) return Tupelpoint is 
      Zeiger : Tupelpoint := null;  
   begin
      for I in reverse Z'range loop
         for J in reverse Z'range loop
            if Matrix(I,J) then
               Zeiger:=new Tupel'(I,J,Zeiger);
            end if;
         end loop;
      end loop;
      return Zeiger;
   end Erzeugemenge;



   -- Diese Prozedur gibt die bergebene Matrix auf dem Bildschirm aus.
   --  Hierzu wird die Matrix und ein Bezeichner bergeben.
   --  Zuerst wird der Bezeichner ausgegeben, anschlieend die Matrix.
   --  Die Funktion unterscheidet zwischen Integer und sonstigen Typen,
   --  da hierbei der Spaltenabstand der Matrixkopfzeile variiert.
   --  Die Wahrheitswerte der Matrix werden hier mit
   --  "0" fr "FALSE" und "1" fr "TRUE" ausgegeben.
   procedure Druckematrix (M1 : Z_Matrix; Bezeichner : String) is 
      Leerzeichen : Boolean := False;  
   begin
      New_Line(2);
      Put(Bezeichner);
      New_Line(2);

      if Z'Image(Z'First)=Integer'Image(1) then
         Leerzeichen:=True;
      end if;

      if Leerzeichen=False then
         Put(" ");
      end if;

      Put("     ");
      for I in Z'range loop
         Put(Z'Image(I));
         if Leerzeichen then
            Put(" ");
         end if;
      end loop;
      New_Line(2);
      for I in Z'range loop
         Put(Z'Image(I));
         Put(" :");
         for J in Z'range loop
            if M1(I,J) then
               Put("  1");
            else
               Put("  0");
            end if;
         end loop;
         New_Line;
      end loop;
      New_Line;
   end Druckematrix;


   -- Diese Prozedur gibt die Menge der Tupel auf dem Bildschirm aus.
   --  Hierzu wird ihr ein Zeiger auf eine Liste von Tupeln bergeben.
   --  Diese Liste wird elementweise durchlaufen und der Inhalt 
   --  ausgegeben.
   procedure Druckemenge (Point : Tupelpoint) is 
      Zeiger : Tupelpoint := Point;  
      Menge  : String     := "";  
   begin
      New_Line;
      Put("   M:={");
      loop
         exit when Zeiger = null;
         Put("("& Z'Image(Zeiger.Pos1)
            &","& Z'Image(Zeiger.Pos2)& ");");
         Zeiger:=Zeiger.Rest;
      end loop;
      Put("}");
      Put(Menge);
      New_Line;
      null;
   end Druckemenge;


end Toadi;